unit main;

// (c)2007 by Paul TOTH <tothpaul@free.fr>
// http://tothpaul.free.fr

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, OpenGL, Snake;

type
  TGLSize=record
   cx:single;
   cy:single;
  end;

  TMainForm = class(TForm)
    procedure FormResize(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
    fDC:HDC;
    fGL:cardinal;
    Snake:TSnake;
    MouseX,MouseY:integer;
    rx,ry:single;
    px,py:single;
    StartTicks:cardinal;
    DeltaTicks:cardinal;
    PauseTicks:cardinal;
    Depth:single;
    Pause:boolean;
    Metrics:array of TGlyphMetricsFloat;
    CamMode:char;
    OldMode:char;
    DieMode:char;
    Mirror :boolean;
    procedure WMEraseBkGnd(var Msg:TMessage); message WM_ERASEBKGND;
    procedure RenderWall;
    procedure RenderCells;
    procedure RenderTrap;
    procedure RenderBlocks;
    procedure Restart;
    function TextSize(const Str:string):TGLSize;
    procedure TextFX(Top:single; const Str:string; Ticks:cardinal);
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}
{
 La fonction rotate ajuste l'angle "r" pour atteindre l'angle "t" suivant le facteur "f"
 Il est ncessaire de tester dans quel sens il faut tourner pour prendre le plus cours chemin
}
function rotate(r,f,t:single):single;
var
 a,b:single;
begin
 a:=t-r;
 b:=t+360-r; if abs(b)<abs(a) then a:=b;
 b:=t-360-r; if abs(b)<abs(a) then a:=b;
 Result:=FixAngle(r+f*a);
end;

{
 le texte est affich en utilisant une "call list" gnre par la fonction wglUseFontOutlines
 celle-ci est cre de tel sorte que la valeur ascii d'un caractre donne directement son index dans la call list
}
procedure glTextOut2(const s:string);
begin
 glPushMatrix;
  glCallLists(Length(s),GL_UNSIGNED_BYTE,@s[1]);
 glPopMatrix;
end;

{
 cette fonction permet d'affiche deux fois un texte avec un dcalage pour donner un effet 3D
}
procedure glTextOut(const s:string);
begin
 glPushMatrix;
  glColor3f(0.5,0.5,0.5);
  glTextOut2(s);
  glTranslatef(0,0,0.25);
  glColor3f(1,1,1);
  glTextOut2(s);
 glPopMatrix;
end;

{
 Initialisation du tout sur la cration de la fiche
}
procedure TMainForm.FormCreate(Sender: TObject);
var
 pfd:TPIXELFORMATDESCRIPTOR;
 pixelformat:integer;
 rgb:cardinal;
begin
// il nous faut un context de device
 fDC:=GetDC(Handle);
 if fDC=0 then RaiseLastOSError;
// nous allons indiquer  Windows les caractristiques de notre fentre OpenGL
 FillChar(pfd,SizeOf(pfd),0);
 pfd.nSize       := sizeof(pfd);
 pfd.nVersion    := 1;
// on dessine directement sur une fiche (et non dans un BITMAP), avec un double buffer
 pfd.dwFlags     := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER;
 pfd.iLayerType  := PFD_MAIN_PLANE;
// l'image est en RGBA 32bits
 pfd.iPixelType  := PFD_TYPE_RGBA;
 pfd.cColorBits  := 32;
// le ZBuffer utilise 24 bits
 pfd.cDepthBits  := 24; // 32 pose des pb sur certaines config ?!
 pixelformat:=ChoosePixelFormat(fDC, @pfd);
 if PixelFormat=0 then RaiseLastOSError;
 if not SetPixelFormat(fDC, pixelformat, @pfd) then RaiseLastOSError;
// maintenant on peut initialiser OpenGL
 fGL:=wglCreateContext(fDC);
// et fixer le context courant (le seul dans cette application)
 wglMakeCurrent(fDC,fGL);
// la couleur d'effacement est calque sur celle de la fiche Delphi
 rgb:=ColorToRGB(Color);
 glClearColor((rgb and $FF)/255,((rgb shr 8) and $FF)/255,((rgb shr 16) and $FF)/255,0);
// valeur d'effacement du ZBuffer
 glClearDepth(1);
// autoriser le test ZBuffer
 glEnable(GL_DEPTH_TEST);
// grer les faces caches
 glEnable(GL_CULL_FACE);
// cration d'une "call list" avec la police (True Type) de la fiche
 SelectObject(fDC,Font.Handle);
// on garde une information sur la taille de chaque caractre
 SetLength(Metrics,220);
// on prend les caractres de #32  #252 en basant la call list sur 32
 if not wglUseFontOutlines(fDC,32,220,32,0,0,WGL_FONT_POLYGONS,@Metrics[0]) then RaiseLastOSError;
// forcer le resize
 FormResize(Self);
// crer le serpent
 Snake:=TSnake.Create;
// fixer les modes par dfaut
 CamMode:='F';
 Mirror:=True;
 Restart;
end;

{ fin de jeu }
procedure TMainForm.FormDestroy(Sender: TObject);
begin
 wglMakeCurrent(fDC,0);
 wglDeleteContext(fGL);
 ReleaseDC(Handle,fDC);
end;

{ retourne la taille d'un texte sous OpenGL }
function TMainForm.TextSize(const Str:string):TGLSize;
var
 i:integer;
 c:integer;
 t:single;
begin
 Result.cx:=0;
 Result.cy:=0;
 for i:=1 to Length(Str) do begin
  c:=ord(Str[i])-32;
  Result.cx:=Result.cx+Metrics[c].gmfCellIncX;
  t:=Metrics[c].gmfBlackBoxY;
  if t>Result.cy then Result.cy:=t;
 end;
end;

{ dbut de partie }
procedure TMainForm.Restart;
begin
 Snake.Reset;
 StartTicks:=GetTickCount;
 DeltaTicks:=StartTicks;
 OldMode:=CamMode;
 CamMode:='G';
end;

{ OpenGL redessine tout la fiche, on vite ici de grer le background }
procedure TMainForm.WMEraseBkGnd(var Msg:TMessage);
begin
 Msg.Result:=1;
end;

{ ajuster le rendu OpenGL quand la fiche est redimensionne }
procedure TMainForm.FormResize(Sender: TObject);
begin
 glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(45,ClientWidth/ClientHeight,0.1,1000);
 glMatrixMode(GL_MODELVIEW);
  glViewport(0, 0, ClientWidth, ClientHeight);
 Invalidate;
end;

{ c'est ici que tout se passe }
procedure TMainForm.FormPaint(Sender: TObject);
var
 t:cardinal;
 f,tx,ty,dx,dy:single;
begin
{ on efface les pixels et le Z Buffer }
 glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
{ reset de la matrice de projection }
 glLoadIdentity;
{ compteur de temps }
 t:=GetTickCount;
{ position de la camra en mode Global }
 if (CamMode='G') then begin
  if (StartTicks>0) then begin
  // pendant le dbut de partie, appliquer un mouvement de camra
   tx:=(t-StartTicks)/100-50;
   ty:=60-60*sin((t-StartTicks)/3000);
  end else begin
  // par la suite, la position est fixe
   tx:=0;
   ty:=0;
  end;
 end;

 // pendant le dbut de partie...
 if StartTicks>0 then begin
 // calcul de la position de la dalle par laquelle entre le serpent
  Depth:=(t-StartTicks)/50;
 // condition de fin d'animation de la dalle
  if Depth>50 then begin
   Depth:=0;
   StartTicks:=0;
   Randomize;
   Snake.MoveFrog; // placer la premire grenouille
  end else
 //  mise chemin, faire remonter la dalle
  if Depth>25 then Depth:=50-Depth else
  if Depth>20 then CamMode:=OldMode; // revenir dans l'tat avec Restart
 end;

 // animer le serpent si on est pas en pause
 if not Pause then Snake.Animate(t-DeltaTicks);

 // dterminer la position de la camra en mode follow
 if (CamMode='F') then begin
  tx:=Snake.Head.RotY;
  ty:=60;
  with Snake.Head.Pos do begin
   dx:=-x;
   dy:=-z;
  end;
 end else begin
 // sinon recenter la camra  l'cran
  dx:=0;
  dy:=0;
 end;

 // A part en mode souris, ajuster la position de la camra en fonction du temps
 if CamMode<>'M' then begin
  f:=(t-DeltaTicks)/250;
  rx:=rotate(rx,f,tx);
  ry:=rotate(ry,f,ty);
  px:=px+f*(dx-px);
  py:=py+f*(dy-py);
 end;

 DeltaTicks:=t;

// positionner la camra
 glTranslatef(0,10,-150+4*ry-300);
 glRotatef(-ry+80,1,0,0);
 glRotatef(-rx,0,1,0);
 glTranslatef(px,0,py);

 if Mirror then begin
 // dessin du miroir
  glPushMatrix;
   glTranslatef(10,-5,10);
  // on retourner le serpent
   glScalef(1,-1,1);
  // et on inverse le test de face cache pour le remettre dans le bon sens
   glCullFace(GL_FRONT);
  // la couleur est diminue pour donner l'impression du reflet
   Snake.Render(True);
   Snake.RenderFrog;
  // on revient en mode normal
   glCullFace(GL_BACK);
  glPopMatrix;
 // placer les cellules dans le Z Buffer (sans les dessiner)
  glClear(GL_DEPTH_BUFFER_BIT);
  RenderCells;
 // effacer l'cran sous les cellules
 // le RenderCells ci-dessus va "protger" le reflet du serpent sur les dalles, tout le reste est remis en noir
 // on ne modifie pas le ZBuffer, il est ici juste question d'effacer une partie de l'image
  glColor3f(0,0,0);
  glDepthMask(FALSE);
  glBegin(GL_QUADS);
   glVertex3f(-8*20,-1,+8*20);
   glVertex3f(+8*20,-1,+8*20);
   glVertex3f(+8*20,-1,-8*20);
   glVertex3f(-8*20,-1,-8*20);
  glEnd;
  glDepthMask(TRUE);
 // ajouter la case blanche (plus bas), elle est dessine  part car plus basse que les autres (le rectangle ci-dessus l'effacerait)
 end else begin
 // sans mirroir, il suffit de "rserver" les cellules, la encore c'est uniquement le ZBuffer qui est affect, la case reste de la couleur du fond (noir)
  RenderCells;
 end;
 if StartTicks>0 then RenderTrap;

 // dessin des murs, 4 fois le mme avec une rotation  90 entre chaque rendu
 RenderWall; glRotatef(90,0,1,0);
 RenderWall; glRotatef(90,0,1,0);
 RenderWall; glRotatef(90,0,1,0);
 RenderWall; glRotatef(90,0,1,0);

 // dessin des colonnes sous les dales
 RenderBlocks;

 // dessin du serpent et de la grenouille
 glColor3f(1,1,1);
 glTranslatef(10,5,10);
 Snake.Render(False);
 Snake.RenderFrog;

 // en dbut de partie, ajouter le texte "GET READY !"
 if StartTicks>0 then begin
  TextFX(0,'GET READY !',t-StartTicks);
 end;

 // en fin de partie, ajouter le score et "GAME OVER !"
 if Snake.Die>0 then begin
  if DieMode=#0 then begin
   DieMode:=CamMode;
   CamMode:='G';
  end;
  TextFX(2,'SCORE '+IntToStr(Snake.Score),Snake.Die);
  TextFX(0,'GAME OVER !',Snake.Die);
 end;

 // en mode pause affiche le texte "PAUSE"
 if Pause then begin
  if PauseTicks=0 then PauseTicks:=t;
 // sauf en mode Souris (pour les screen shoot)
  if CamMode<>'M' then TextFX(0,'PAUSE',t-PauseTicks);
 end else begin
  PauseTicks:=0;
 end;

 // relancer la partie 2 secondes aprs le GAME OVER
 if Snake.Die>2000 then begin
  CamMode:=DieMode;
  DieMode:=#0;
  Restart;
 end;

 // swaper les buffers vidos
 glFlush;
 SwapBuffers(fDC);

 // invalider la fiche pour que le Paint relancer en boucle :)
 InvalidateRect(Handle,nil,False);
end;

// dessin d'un mur, on utilise tout simplement un rectangle bicolor dont OpenGL fait le dgrad
procedure TMainForm.RenderWall;
const
 k1=160;
 k2=50;
var
 rgb:cardinal;
begin
 rgb:=ColorToRGB(Color);
 glBegin(GL_QUADS);
  glColor3f(114/255,234/255,1.0);
  glVertex3f(-k1,-k2,-k1);
  glVertex3f(+k1,-k2,-k1);
  glColor3f((rgb and $FF)/255,((rgb shr 8) and $FF)/255,((rgb shr 16) and $FF)/255);
  glVertex3f(+k1,k2,-k1);
  glVertex3f(-k1,k2,-k1);
 glEnd;
end;

// rserve les dales dans le ZBuffer (pas de rendu, c'est la couleur de fond - ou le reflet - qui reste affiche)
procedure TMainForm.RenderCells;
const
 k =9;
 k2=50;
var
 m:array[0..15] of single;
 x,y:integer;
begin
// ne rien dessiner
 glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
 glGetFloatV(GL_MODELVIEW_MATRIX,@m);
 for x:=0 to 15 do begin
  glTranslatef(-20*8+10+20*x,0,-20*8+10);
  for y:=0 to 15 do begin
   if (StartTicks>0)and(x=8)and(y=8) then begin
   // ignorer cette case, elle est blanche
   end else begin
    glBegin(GL_QUADS);
     glVertex3f(+k,0,-k);
     glVertex3f(-k,0,-k);
     glVertex3f(-k,0,+k);
     glVertex3f(+k,0,+k);
    glEnd;
   end;
   glTranslatef(0,0,20);
  end;
  glLoadMatrixf(@m);
 end;
// rautoriser le dessin
 glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
end;

// dessin de la trape blanche
procedure TMainForm.RenderTrap;
const
 k=9;
begin
 glPushMatrix;
  glTranslatef(10,-Depth,10);
  glColor3f(1,1,1);
  glBegin(GL_QUADS);
   glVertex3f(+k,0,-k);
   glVertex3f(-k,0,-k);
   glVertex3f(-k,0,+k);
   glVertex3f(+k,0,+k);
  glEnd;
 glPopMatrix;
end;

// dessin de la base de dales
procedure TMainForm.RenderBlocks;
const
 k =9;
 k2=50;
var
 m:array[0..15] of single;
 x,y:integer;
 i:integer;
 z:single;
begin
 glGetFloatV(GL_MODELVIEW_MATRIX,@m);
// pour chaque dale, dessiner sa base en quatre rectangle  90
 for x:=0 to 15 do begin
  glTranslatef(-20*8+10+20*x,0,-20*8+10);
  for y:=0 to 15 do begin
   if (StartTicks>0)and(x=8)and(y=8) then z:=-Depth else z:=0;
   for i:=0 to 3 do begin
    glBegin(GL_QUADS);
     glColor3f(114/255,234/255,1.0);
     glVertex3f(-k,z,-k);
     glColor3f(100/255,200/255,1.0);
     glVertex3f(+k,z,-k);
     glColor3f(1,1,1);
     glVertex3f(+k,z-k2,-k);
     glVertex3f(-k,z-k2,-k);
    glEnd;
    glRotatef(90,0,1,0);
   end;
   glTranslatef(0,0,20);
  end;
  glLoadMatrixf(@m);
 end;
// finalement ajouter un grand rectangle blanc en dessous du tout pour donner un effet lumineux entre les dales
 glColor3f(1,1,1);
 glBegin(GL_QUADS);
  glVertex3f(+20*8,-k2,-20*8);
  glVertex3f(-20*8,-k2,-20*8);
  glVertex3f(-20*8,-k2,+20*8);
  glVertex3f(+20*8,-k2,+20*8);
 glEnd;
end;

// gestion des touches
procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 case Key of
  VK_LEFT : Snake.Move:='L';
  VK_RIGHT: Snake.Move:='R';
  VK_SPACE: Pause:=not Pause;
  VK_ESCAPE: Application.Terminate;
  ord('V'): Snake.Grow;
  ord('G'): CamMode:='G';
  ord('F'): CamMode:='F';
  ord('M'): Mirror:=not Mirror;
 end;
end;

// passer la camra en mode souris
procedure TMainForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 MouseX:=x;
 MouseY:=y;
 OldMode:=CamMode;
 CamMode:='M';
end;

// mouvement souris de la camra
procedure TMainForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
 if ssLeft in Shift then begin
  rx:=FixAngle(rx+(x-MouseX));
  ry:=ry+(y-MouseY);
  if ry>75 then ry:=75;
  if ry< 0 then ry:= 0;
  MouseX:=x;
  MouseY:=y;
  Invalidate;
 end;
end;

// sortir du mode camra
procedure TMainForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 CamMode:=OldMode;
end;

// dessiner un texte qui s'affiche dynamiquement de la droite vers le centre de l'cran
procedure TMainForm.TextFX(Top:single; const Str:string; Ticks:cardinal);
var
 i:integer;
 s:TGLSize;
 f:single;
 x,o,w:single;
begin
// if Ticks>2000 then exit;
 glClear(GL_DEPTH_BUFFER_BIT);
 glLoadIdentity;
 glTranslatef(0,Top-2,-15);
 glColor3f(1,1,1);
 s:=TextSize(Str);
 w:=s.cx;
 x:=-w/2;
 for i:=1 to Length(Str) do begin
  s:=TextSize(Str[i]);
  if Ticks<500 then begin
   f:=x+8-8*(Ticks/500);
   s.cx:=s.cx+8-8*(Ticks/500);
  end else
   f:=x;
  glTranslatef(+f,0,0);
  glTextOut(Str[i]);
  glTranslatef(-f,0,0);
  x:=x+s.cx;
 end;
end;

end.
